#include <avr/io.h>
#include <avr/delay.h>
#include "lcd-routines.h"
 
// Max. Delay bei _delay_us() ist 768 us / F_CPU in MHz. 
// Bei 16 MHz lassen sich max. 48 us. erreichen. Daher
// Aufspaltung in 2 kleinere Delays.
#define DELAY50US() \
           _delay_us(25.0); \
           _delay_us(25.0)
 
#define CLEAR_DISPLAY 0x01
#define CURSOR_HOME   0x02
 
// sendet ein Datenbyte an das LCD
void lcd_data(unsigned char temp1)
{
   unsigned char temp2 = temp1;
 
   temp1 = temp1 >> 4;
   temp1 = temp1 & 0x0F;
   PORTD = temp1 | (1<<PD4); // inkl. PD4 = RS-Leitung setzen
   lcd_enable();
 
   temp2 = temp2 & 0x0F;
   PORTD = temp2 | (1<<PD4); // inkl. PD4 = RS-Leitung setzen
   lcd_enable();
   
   DELAY50US();
}
 
// sendet einen Befehl an das LCD
// wie lcd_data, nur ohne RS zu setzen
void lcd_command(unsigned char temp1)
{
   unsigned char temp2 = temp1;
 
   temp1 = temp1 >> 4;   // oberes Nibble holen
   temp1 = temp1 & 0x0F; // maskieren
   PORTD = temp1;        // setzen
   lcd_enable();
 
   temp2 = temp2 & 0x0F; // unteres Nibble holen und maskieren
   PORTD = temp2;        // setzen
   lcd_enable();
   
   DELAY50US();
}
 
// erzeugt den Enable-Puls
void lcd_enable(void)
{
   PORTD |= (1<<PD5);
   asm("nop");
   asm("nop");
   asm("nop");
   PORTD &= ~(1<<PD5);
}
 
// Initialisierung: 
// Muss ganz am Anfang des Programms aufgerufen werden.
// Port D (DDRD) muss bereits als Ausgang eingestellt sein.
void lcd_init(void)
{
   unsigned char temp3 = 50;
 
   while(temp3 != 0)
   {
      _delay_ms(5.0);  // 5ms = 250ms warten
      temp3 = temp3 - 1;
   }
 
   // muss 3mal hintereinander gesendet werden zur Initialisierung
   PORTD = 0x03;       // 0b00000011  
   lcd_enable();
   PORTD = 0x03;
   lcd_enable();
   PORTD = 0x03;
   lcd_enable();
 
   // 4bit-Modus einstellen
   PORTD = 0x02;       // 0b00000010  
   lcd_enable();
   _delay_ms(5.0);
 
   // 4Bit / 2 Zeilen / 5x8
   lcd_command(0x28);  // 0b00101000        
 
   // Display ein / Cursor aus / kein Blinken
   lcd_command(0x0C);  // 0b00001100        
 
   // inkrement / kein Scrollen
   lcd_command(0x04);  // 0b00000100
}
 
// Sendet den Befehl zur Lschung des Displays
void lcd_clear(void)
{
   lcd_command(CLEAR_DISPLAY);
   _delay_ms(5.0);
}
 
// Sendet den Befehl: Cursor Home
void lcd_home(void)
{
   lcd_command(CURSOR_HOME);
   _delay_ms(5.0);
}

// Ich habe kein Hardwareaufbau und kann das Programm nicht testen 
// deshalb bitte ich jemanden kurz das Programm zu verifizieren.
 
// Floating point bergabe an _delay_us(); _delay_ms(); korrigiert 
// Pinbelegung ber define's einstellbar
// Bitmaskierung in den Funktion verbessert
 
#include <avr/io.h>
#include <util/delay.h>
#include "lcd-routines.h"
 
// Max. Delay bei _delay_us() ist 768 us / F_CPU in MHz. 
// Bei 16 MHz lassen sich max. 48 us. erreichen. Daher
// Aufspaltung in 2 kleinere Delays.
#define DELAY50US() _delay_us(25); _delay_us(25)
 
 
// LCD Befehle
 
#define CLEAR_DISPLAY 0x01
#define CURSOR_HOME   0x02
 
 
// Pinbelegung fr das LCD
 
#define PORT_LCD			PORTD
#define LCD_RS				4
#define	LCD EN				5
 
 
// sendet ein Datenbyte an das LCD
void lcd_data(unsigned char temp1)
{
   unsigned char temp2 = temp1;
 
   temp1 = temp1 >> 4;
   temp1 = temp1 & 0x0F;
   PORT_LCD &= 0xF0;
   PORT_LCD |= temp1 | (1<<LCD_RS); // inkl. LCD_RS = RS-Leitung setzen
   lcd_enable();
 
   temp2 = temp2 & 0x0F;
   PORT_LCD &= 0xF0;
   PORT_LCD |= temp2 | (1<<LCD_RS); // inkl. LCD_RS = RS-Leitung setzen
   lcd_enable();
   
   DELAY50US();
}
 
// sendet einen Befehl an das LCD
// wie lcd_data, nur ohne RS zu setzen
void lcd_command(unsigned char temp1)
{
   unsigned char temp2 = temp1;
 
   temp1 = temp1 >> 4;   // oberes Nibble holen
   temp1 = temp1 & 0x0F; // maskieren
   PORT_LCD &= 0xF0;
   PORT_LCD |= temp1;        // setzen
   lcd_enable();
 
   temp2 = temp2 & 0x0F; // unteres Nibble holen und maskieren
   PORT_LCD &= 0xF0;
   PORT_LCD |= temp2;        // setzen
   lcd_enable();
   
   DELAY50US();
}
 
// erzeugt den Enable-Puls
void lcd_enable(void)
{
   PORT_LCD |= (1<<LCD_EN);
   asm("nop");
   asm("nop");
   asm("nop");
   PORT_LCD &= ~(1<<LCD_EN);
}
 
// Initialisierung: 
// Muss ganz am Anfang des Programms aufgerufen werden.
// LCD_PORT  (DDRx) muss bereits als Ausgang eingestellt sein.
void lcd_init(void)
{
   unsigned char temp3 = 50;
 
   while(temp3 != 0)
   {
      _delay_ms(5);  // 5ms = 250ms warten
      temp3 = temp3 - 1;
   }
 
   // muss 3mal hintereinander gesendet werden zur Initialisierung
   PORT_LCD &= 0xF0;
   PORT_LCD |= 0x03;       // 0b00000011  
   lcd_enable();
   PORT_LCD |= 0x03;
   lcd_enable();
   PORT_LCD |= 0x03;
   lcd_enable();
 
   // 4bit-Modus einstellen
   PORT_LCD |= 0x02;       // 0b00000010  
   lcd_enable();
   _delay_ms(5);
 
   // 4Bit / 2 Zeilen / 5x8
   lcd_command(0x28);  // 0b00101000        
 
   // Display ein / Cursor aus / kein Blinken
   lcd_command(0x0C);  // 0b00001100        
 
   // inkrement / kein Scrollen
   lcd_command(0x04);  // 0b00000100
}
 
// Sendet den Befehl zur Lschung des Displays
void lcd_clear(void)
{
   lcd_command(CLEAR_DISPLAY);
   _delay_ms(5);
}
 
// Sendet den Befehl: Cursor Home
void lcd_home(void)
{
   lcd_command(CURSOR_HOME);
   _delay_ms(5);
}
